# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, absolute_import

# create a new AC automaton
from aca import Automaton

automaton = Automaton()
automaton['Estonia'] = 'Tallinn'
automaton['Germany'] = 'Berlin'
automaton['Finland'] = 'Helsinki'

# serialize to disk
automaton.save_to_file('myautomaton.bin')

# load from disk
automaton2 = Automaton()
automaton2.load_from_file('myautomaton.bin')

# save / load to binary string
automaton3 = Automaton()
automaton3.load_from_string(automaton.save_to_string())

print (automaton2['Estonia'])
print (automaton3['Germany'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, absolute_import

from aca import Automaton, Match


def test_automaton_with_words():
    auto = Automaton()
    for token in ['he', 'she', 'his', 'hers']:
        auto.add(token)

    expected_all_matches = [Match(1, 4, 'Y'), Match(2, 4, 'Y'), Match(2, 6, 'Y')]
    all_matches = auto.get_matches('ushers', exclude_overlaps=False)
    print (all_matches)
    assert expected_all_matches == all_matches

    expected_nonoverlap_matches = [Match(2, 6, 'Y')]
    nonoverlap_matches = auto.get_matches('ushers', exclude_overlaps=True)
    assert expected_nonoverlap_matches == nonoverlap_matches


def test_with_updating():
    auto = Automaton()
    auto.add('hers')
    matches = auto.get_matches('ushers')
    assert len(matches) == 1
    auto.add('us')
    matches = auto.get_matches('ushers')
    assert len(matches) == 2


def test_with_words():
    auto = Automaton()
    auto.add(['funderbeam'])
    auto.add(['mattermark'])
    auto.add(['500', 'startups'])

    txt = 'funderbeam and mattermark along with 500 startups'.split()
    expected = [Match(0, 1, 'Y'), Match(2, 3, 'Y'), Match(5, 7, 'Y')]
    actual = auto.get_matches(txt)
    assert expected == actual


def test_lemmas():
    auto = Automaton()
    auto.add(['sunlabob'], 'CO')
    auto.add(['renewable'], 'CO')
    lemmas = ['sunlabob', 'renewable', 'energy', 'receive', '$', '2.1', 'million', 'investment']
    print (auto.str())
    matches = auto.get_matches(lemmas)
    assert len(matches) == 2


test_lemmas()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, absolute_import

# create a new AC automaton
from aca import Automaton
map = Automaton()

# use the automaton as a map
map['electrify'] = 'verb'
map['elegant'] = 'adjective'
map['acid'] = 'noun'
map['acidic'] = 'adjective'

# access it like a Python dictionary
print (map['acid'])

# using an invalid key raises a KeyError
#print (map['invalid key'])

# you can use get to provide a default value when key is missing
print (map.get('invalid key', 'default value'))

# NB! Implementation specific special case: empty strings
# denote "missing" values, so you can't use these
map['special'] = ''
#print (map['special'])

# you can delete items
del map['electrify']

# trying to delete a non-existent item raises KeyError
#del map['invalid key']

# iterate items like a dict
print ('items:')
for key, value in map.items():
    print ('{}: {}'.format(key, value))


# you can also iterate prefixes
print ('prefixes:')
for prefix, value in map.prefixes():
    print ('{}: {}'.format(prefix, value))



# Import the library and initiate the automaton
from aca import Automaton
automaton = Automaton()

# add the entities and build the automaton
automaton.add_all(['Funderbeam', 'Funderbeam Data', 'Funderbeam Markets'])
automaton.update_automaton()

# find matches
text = 'Funderbeam Data and Funderbeam Markets are two different products of Funderbeam'
for match in automaton.get_matches(text, exclude_overlaps=False):
    print (match.start, match.end, match.elems)

for match in automaton.get_matches(text, exclude_overlaps=True):
    print (match.start, match.end, match.elems)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, absolute_import

from aca import Automaton
import sys

names = [('janek', 'nice'), ('jaan', 'nice'), ('jaagup', 'ugly'), ('jaanus', 'nice'), ('janis', 'nice')]


def test_items():
    auto = Automaton()
    auto.add_all(names)
    ens, evs = zip(*sorted(names))
    ns, vs = zip(*list(auto.items()))
    ns = [''.join(n) for n in ns]
    assert list(ens) == list(ns)
    assert list(evs) == list(vs)


def test_prefixes():
    auto = Automaton()
    auto.add_all(['jaanus', 'janek', 'janis'])
    prefixes, values = zip(*auto.prefixes())
    prefixes = [''.join(prefix) for prefix in prefixes]
    assert prefixes == ['', 'j', 'ja', 'jaa', 'jaan', 'jaanu', 'jaanus', 'jan', 'jane', 'janek', 'jani', 'janis']




# distutils: language = c++
# distutils: sources = aca/match.cpp aca/node.cpp aca/automaton.cpp
# -*- coding: utf-8 -*-
#from __future__ import unicode_literals, print_function, absolute_import

import six
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.utility cimport pair
import unicodedata

cdef extern from "all.h" namespace "aca":
    cdef cppclass CppMatch:
        CppMatch() except +
        CppMatch(int, int, string) except +
        CppMatch(int, int, char*) except +
        void set_start(int)
        void set_end(int)
        void set_label(string)
        int get_start()
        int get_end()
        string get_label()
        int is_before(const CppMatch&)
        size_t size()

    cdef vector[CppMatch] cpp_remove_overlaps(vector[CppMatch]);

    cdef cppclass CppAutomaton:
        Automaton() except +
        void add(vector[string]&, string)
        void update_automaton()
        bool has_pattern(vector[string]&)
        bool has_prefix(vector[string]&)
        string get_value(vector[string]&)
        vector[CppMatch] get_matches(vector[string]&, bool)
        vector[pair[vector[string], string]] get_patterns_values()
        vector[pair[vector[string], string]] get_prefixes_values()

        # serialization related
        string serialize()
        void serialize_to (string) except +

        @staticmethod
        CppAutomaton* deserialize(string)

        @staticmethod
        CppAutomaton* deserialize_from(string) except +

        string str()


def normalize_unicode(text):
    return unicodedata.normalize('NFC', text)

def encode(txt):
    if isinstance(txt, six.string_types):
        return normalize_unicode(txt).encode('utf-8')
    return txt

def decode(txt):
    if isinstance(txt, six.binary_type):
        return txt.decode('utf-8')
    return txt

def encode_list(lst):
    return [encode(e) for e in lst]

def decode_list(lst):
    return [decode(e) for e in lst]


class Match:

    def __init__(self, start, end, label='Y'):
        self.__start = int(start)
        self.__end = int(end)
        assert self.__start < self.__end
        self.__label = str(label)
        self.__elems = None

    def __eq__(self, other):
        return self.start == other.start and self.end == other.end and self.label == other.label

    def __str__(self):
        return 'Match({},{},{},{})'.format(self.start, self.end, self.elems, self.label)

    def __repr__(self):
        return str(self)

    @property
    def start(self):
        return self.__start

    @property
    def end(self):
        return self.__end

    @property
    def label(self):
        return self.__label

    @property
    def elems(self):
        return self.__elems

    def set_elems(self, elems):
        self.__elems = elems


cdef vector[CppMatch] matches_to_cppmatches(matches):
    cdef vector[CppMatch] vec
    cdef CppMatch cppmatch;
    vec.reserve(len(matches))
    for match in matches:
        cppmatch.set_start(match.start)
        cppmatch.set_end(match.end)
        cppmatch.set_label(encode(match.label))
        vec.push_back(cppmatch)
    return vec

cdef cppmatches_to_matches(vector[CppMatch] cppmatches):
    result = [None]*cppmatches.size()
    for i in range(cppmatches.size()):
        result[i] = Match(cppmatches[i].get_start(), cppmatches[i].get_end(), decode(cppmatches[i].get_label()))
    return result

def remove_overlaps(matches):
    cdef vector[CppMatch] cppmatches;
    cdef vector[CppMatch] cppresult;
    cppmatches = matches_to_cppmatches(matches)
    cppresult = cpp_remove_overlaps(cppmatches)
    return cppmatches_to_matches(cppresult)


cdef class Automaton:
    """ Aho-Corasick keyword tree + automaton. """
    cdef CppAutomaton* cpp_automaton

    def __cinit__(self):
        self.cpp_automaton = new CppAutomaton()

    def __dealloc__(self):
        del self.cpp_automaton

    def load_from_file(self, fnm):
        cdef CppAutomaton* new_cpp_automaton = self.cpp_automaton.deserialize_from(encode(fnm))
        del self.cpp_automaton
        self.cpp_automaton = new_cpp_automaton

    def load_from_string(self, binstring):
        cdef CppAutomaton* new_cpp_automaton = self.cpp_automaton.deserialize(binstring)
        del self.cpp_automaton
        self.cpp_automaton = new_cpp_automaton

    def add(self, pattern, value='Y'):
        self.cpp_automaton.add(encode_list(pattern), encode(value))

    def add_all(self, patterns):
        for pattern in patterns:
            if isinstance(pattern, tuple):
                pattern, value = pattern
                self.add(pattern, value)
            else:
                self.add(pattern)

    def update_automaton(self):
        self.cpp_automaton.update_automaton()

    def has_pattern(self, pattern):
        return self.cpp_automaton.has_pattern(encode_list(pattern))

    def has_prefix(self, prefix):
        return self.cpp_automaton.has_prefix(encode_list(prefix))

    def get_matches(self, text, exclude_overlaps=True):
        matches = self.cpp_automaton.get_matches(encode_list(text), exclude_overlaps)
        results = cppmatches_to_matches(matches)
        for match in results:
            match.set_elems(text[match.start:match.end])
        return results

    def items(self):
        cdef vector[pair[vector[string], string]] vec = self.cpp_automaton.get_patterns_values()
        for idx in range(vec.size()):
            yield decode_list(vec[idx].first), decode(vec[idx].second)

    def prefixes(self):
        cdef vector[pair[vector[string], string]] vec = self.cpp_automaton.get_prefixes_values()
        for idx in range(vec.size()):
            yield decode_list(vec[idx].first), decode(vec[idx].second)

    '''def keys(self):
        for key, value in self.items():
            yield tuple(key)

    def values(self):
        for key, value in self.items():
            yield value

    def __iter__(self):
        return self.keys()'''

    def save_to_file(self, fnm):
        self.cpp_automaton.serialize_to(encode(fnm))

    def save_to_string(self):
        return self.cpp_automaton.serialize()

    def __getitem__(self, pattern):
        value = decode(self.cpp_automaton.get_value(encode_list(pattern)))
        if len(value) == 0:
            raise KeyError(pattern)
        return value

    def get(self, pattern, default=None):
        try:
            return self[pattern]
        except KeyError as exc:
            return default

    def __setitem__(self, pattern, value):
        self.add(pattern, value)

    def set(self, pattern, value=''):
        self.add(pattern, value)

    def __delitem__(self, pattern):
        value = self[pattern] # if value does not exist, raise an exception
        self.add(pattern, '')

    def __contains__(self, pattern):
        value = self.get(pattern, None)
        return value is not None

    def str(self):
        return decode(self.cpp_automaton.str())




import setuptools
import platform
from distutils.core import setup, Extension
from Cython.Build import cythonize
from distutils.command.sdist import sdist as _sdist
from setuptools import find_packages

# this is for making the package documentation look nice on PyPi
try:
    import pypandoc
    LONG_DESCRIPTION = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    LONG_DESCRIPTION = open('README.md').read()


NAME = "aca"
VERSION = '0.9'
DESCRIPTION = 'Aho-Corasick automaton implementation in C++'
AUTHOR = 'Timo Petmanson @Funderbeam'
AUTHOR_EMAIL = 'tpetmanson@gmail.com'
LICENSE = 'GPLv3'
URL = 'https://github.com/tpetmanson/aca'

CLASSIFIERS = [
    'Intended Audience :: Developers',
    'Topic :: Text Processing',
    'Topic :: Text Processing :: Linguistic',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5'
]


REQUIREMENTS = [
    'Cython>=0.25.0,<0.26',
    'wheel>0.25.0',
    'six',
    'pytest']

EXTRA_ARGS = ['-std=c++11']

osname = platform.system().lower()
if 'linux' in osname:
    pass
if 'darwin' in osname:
    EXTRA_ARGS.append('-mmacosx-version-min=10.9')
if 'windows' in osname:
    pass # TODO

EXTENSIONS = [
    Extension('aca.aca_cpp',
              sources=['aca/aca_cpp.pyx'],
              language='c++',
              extra_compile_args=EXTRA_ARGS,
              extra_link_args=['-std=c++11'])]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    setup_requires=REQUIREMENTS,
    install_requires=REQUIREMENTS,
    packages=find_packages(),
    include_package_data=True,

    ext_modules=cythonize(EXTENSIONS)
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, absolute_import

"""
Example/test script that detects names from a longer text
"""
from aca import Automaton, Match

NAMES = '''
Viktor Mikhaylovich Afanasyev
Vladimir Aksyonov
Aleksandr Pavlovich Aleksandrov
Ivan Anikeyev s
Anatoly Artsebarsky
Yuri Artyukhin 
Oleg Atkov
Toktar Aubakirov
Sergei Avdeyev
Nikolai Budarin
Vladimir Dezhurov
Aleksandr Balandin
Yuri Baturin
Pavel Belyayev 
Georgi Beregovoi 
Anatoly Berezovoy 
Valentin Bondarenko 
Andrei Borisenko
Nikolai Budarin
Valery Bykovsky
Vladimir N. Dezhurov
Georgiy Dobrovolskiy 
Lev Dyomin 
Vladimir Dzhanibekov
Konstantin Feoktistov 
Valentin Filatyev 
Anatoly Filipchenko
Yuri Gagarin 
Yuri Gidzenko
Yuri Glazkov 
Viktor Gorbatko
Georgi Grechko
Aleksei Gubarev
Aleksandr Kaleri
Sergei Krikalev
Aleksandr Ivanchenkov
Anatoli Ivanishin
Aleksandr Kaleri
Yevgeny Khrunov 
Leonid Kizim 
Pyotr Klimuk
Vladimir Komarov 
Yelena V. KondakovaSymbol venus.svg
Dmitri Kondratyev
Oleg Kononenko
Mikhail Korniyenko
Valery Korzun
Oleg Kotov
Vladimir Kovalyonok
Konstantin Kozeyev
Sergei Krikalev
Valeri Kubasov
Aleksei Leonov
Aleksandr Laveykin
Vasili Lazarev 
Aleksandr Lazutkin
Valentin Lebedev
Aleksei Leonov
Anatoli Levchenko 
Yuri Lonchakov
Vladimir Lyakhov
Yuri Lonchakov
Oleg Makarov 
Yuri Malenchenko
Yury Malyshev 
Gennadi Manakov
Musa Manarov
Alexander Misurkin
Boris Morukov 
Talgat Musabayev
Gennady Padalka
Grigori Nelyubov 
Andriyan Nikolayev 
Oleg Novitski
Yuri Onufrienko
Aleksey Ovchinin
Gennady Padalka
Viktor Patsayev 
Aleksandr Poleshchuk
Valeri Polyakov
Leonid Popov
Pavel Popovich 
Sergei Revin
Roman Romanenko
Yuri Romanenko
Valery Rozhdestvensky
Nikolai Rukavishnikov 
Sergei Ryazanski
Valery Ryumin
Leonid Popov
Svetlana Savitskaya
Aleksandr Serebrov
Aleksandr Samokutyayev
Gennadi Sarafanov 
Viktor Savinykh
Svetlana SavitskayaSymbol venus.svg
Aleksandr Serebrov 
Yelena SerovaSymbol venus.svg
Vitali Sevastyanov 
Yuri Shargin
Salizhan Sharipov
Vladimir Shatalov
Anton Shkaplerov
Georgi Shonin 
Oleg Skripochka
Aleksandr Skvortsov
Anatoly Solovyev
Vladimir Solovyov
Gennadi Strekalov 
Maksim Surayev
Valentina Tereshkova
Yevgeni Tarelkin
Valentina TereshkovaSymbol venus.svg, First woman in space.
Gherman Titov 
Vladimir Titov
Valeri Tokarev
Sergei Treshchov
Vasili Tsibliyev
Mikhail Tyurin
Yuri Usachov
Vladimir Vasyutin 
Aleksandr Viktorenko
Pavel Vinogradov
Igor Volk
Alexander Volkov
Sergei Aleksandrovich Volkov
Vladislav Volkov 
Boris Volynov
Fyodor Yurchikhin
Sergei Vozovikov 
Boris Yegorov 
Aleksei Yeliseyev
Fyodor Yurchikhin
Dmitri Zaikin 
Sergei Zalyotin
Vitali Zholobov
Vyacheslav Zudov
Marcos Pontes
Guy Laliberté
Chris Hadfield
Robert Thirsk
Joseph M. Acaba
Michael R. Barratt
Ken Bowersox
Daniel C. Burbank
Leroy Chiao
Timothy Creamer
Michael Fincke
Michael Foale
Michael E. Fossum
Ronald J. Garan, Jr.
Richard Garriott
Michael S. Hopkins
Scott Kelly
Kjell N. Lindgren
Michael Lopez-Alegria
Ed Lu
Thomas Marshburn
Richard Mastracchio
William S. McArthur
Karen Nyberg
Gregory Olsen
Donald Pettit
John L. Phillips
William Shepherd
Charles Simonyi
Norman Thagard
Dennis Tito
Terry W. Virts
Shannon WalkerSymbol
Douglas H. Wheelock
Peggy WhitsonSymbol
Jeffrey Williams
Sunita WilliamsSymbol
Barry E. Wilmore
Gregory R. Wiseman
Toyohiro Akiyama
Akihiko Hoshide
Soichi Noguchi
Satoshi Furukawa
Koichi Wakata
Kimiya Yui
'''

TEXT = ' who was a very good friend of '.join(NAMES.strip().splitlines())
NAMES = [name.split() for name in NAMES.strip().splitlines()]


def test_names():
    auto = Automaton(NAMES)
    auto.add_all(NAMES)
    print (auto)

    matches = set(' '.join(match.elems) for match in auto.get_matches(TEXT.split()))
    names = set(' '.join(name) for name in NAMES)

    assert names == matches




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, absolute_import

import pytest
from aca import Automaton

def test_has_pattern():
    automaton = Automaton()
    automaton.add('himalaya')

    print (automaton)

    assert automaton.has_prefix('him')
    assert automaton.has_prefix('himalaya')
    assert not automaton.has_prefix('himalayas')
